/*
** Module   :UNDO.CPP
** Abstract :UNDO processor
**
** Copyright (C) Sergey I. Yevtushenko
**
** Log: Mon  15/03/1998     Created
**      Fri  13/06/2003     Removed block processing code (now in block.cpp)
*/

#include <string.h>

#include <buffer.h>
#include <version.h>

//----------------------------------------------------------------------
// Undo processing routines
//
//----------------------------------------------------------------------

void Buffer::track_beg()
{
    tracking = 1;
}

void Buffer::track_end()
{
    if(!tracking)
        return;
    if(!track_head)
    {
        tracking   = 0;
        return;
    }

    trackinfo *head = track_head; //Save track
    track_head = undobuff;        //spoof 'track'

    track(opAction, head);

    undobuff   = track_head;
    track_head = 0;
    tracking   = 0;
    undo_count++;
}

void Buffer::track_cancel()
{
    if(track_head) //cleanup here
    {
        while(track_head)
        {
            switch(track_head->op)
            {
                case opInsBlock:
                    delete (Buffer *)track_head->arg1;
                    break;
                case opRestoreLine:
                case opInsLine:
                    Free(track_head->arg1);
                    break;
//                case opRestoreOrder:
//                    delete (char*)(track_head->arg1);
            }
            trackinfo *tmp = track_head;
            track_head = track_head->next;
            delete tmp;
        }
    }
    tracking = 0;
}

void Buffer::track(int op, void *arg1, void *arg2)
{
    if(!tracking)
    {
        if(op == opInsBlock)
            delete (Buffer *)arg1;
        return;
    }

    trackinfo *item = new trackinfo;

    item->next = track_head;
    item->op   = op;
    item->arg1 = arg1;
    item->arg2 = arg2;
    track_head = item;
    switch(op)
    {
        case opRestoreLine:
        case opInsLine:
            // In this cases arg1 contains pointer to line,
            // we need a copy of them

            item->arg1 = new Line(PLine(arg1));
            break;
    }
}

void Buffer::undo(Rect&)
{
    trackinfo *item = undobuff;
    int need_recalc = 0;

    if(item) //Undo buffer always contains opAction items
    {
        undobuff = item->next;

        trackinfo *action = (trackinfo *)item->arg1;

        while(action)
        {
            // Do action
            switch(action->op)
            {
                case opCurRow:
                    cur_row = (int)action->arg1;
                    break;

                case opCurCol:
                    cur_col = (int)action->arg1;
                    break;

                case opStartRow:
                    start_row = (int)action->arg1;
                    break;

                case opStartCol:
                    start_col = (int)action->arg1;
                    break;

                case opColBlock:
                    col_block = (int)action->arg1;
                    break;

                case opChanged:
                    changed = (int)action->arg1;
                    if(changed)
                        saved = 0;
                    break;

                case opInsMode:
                    ins_mode = (int)action->arg1;
                    break;

                case opHiliting:
                    hiliting = (int)action->arg1;
                    need_recalc = 1;
                    break;

                case opMarking:
                    mark_state = (int)action->arg1;
                    break;

                case opDelLine:
                    Free(Remove((unsigned)action->arg1));
                    need_recalc = 1;
                    break;

                case opInsBlock:
                    {
                        Buffer* tmp_buf = (Buffer *)action->arg1;
                        int start_pos   = (int)action->arg2;
                        move_items(tmp_buf, start_pos);
                        delete tmp_buf;
                        need_recalc = 1;
                    }
                    break;

                case opInsLine:
                    At(action->arg1, (unsigned) action->arg2);
                    need_recalc = 1;
                    break;

                case opRestoreLine:
                    Free(Remove((unsigned)action->arg2));
                    At(action->arg1,(unsigned)action->arg2);
                    need_recalc = 1;
                    break;

                case opMarkPos:
                    old_abs_col = (int)action->arg1;
                    old_abs_row = (int)action->arg2;
                    break;

                case opInsChar:
                    abs_line()->ins_char(chr_in((int)action->arg1), abs_col());
                    need_recalc = 1;
                    break;

                case opDelChar:
                    abs_line()->del_char(abs_col());
                    need_recalc = 1;
                    break;

            }

            trackinfo *temp = action->next;

            delete action;
            action = temp;
        }
        delete item;
        undo_count--;

        if(undo_count < 0)
            undo_count = 0;
    }

    if(hiliting && need_recalc)
        fill_hiliting(0, ST_INITIAL);
}

int Buffer::get_undo_count()
{
    return undo_count;
}

void Buffer::clear_undobuff()
{
    while(undobuff)
    {
        trackinfo *item = undobuff;

        if(item) //Undo buffer always contains opAction items
        {
            undobuff = item->next;
            trackinfo *action = (trackinfo *)item->arg1;

            while(action)
            {
                // Cleanup action item

                switch(action->op)
                {
                    case opRestoreLine:
                    case opInsLine:
                        Free(action->arg1);
                        break;
                }

                trackinfo *temp = action->next;

                delete action;
                action = temp;
            }
            delete item;
        }
    }
    undo_count = 0;
}

